;; Generated package description from orgalist.el  -*- no-byte-compile: t -*-
(define-package "orgalist" "1.15.0.20240617.180607" "Manage Org-like lists in non-Org buffers" '((emacs "26.1")) :commit "6b97873be78dba574f4753104c8f2da3c8bfb5ee" :url "https://elpa.gnu.org/packages/orgalist.html" :authors '(("Nicolas Goaziou" . "mail@nicolasgoaziou.fr")) :maintainer '("Nicolas Goaziou" . "mail@nicolasgoaziou.fr") :keywords '("convenience"))
